@@ -4,7 +4,7 @@
  *
  * Copyright (c) 1994-2000 Eric Youngdale, Peter MacDonald, 
  *				David Engel, Hongjiu Lu and Mitch D'Souza
- * Copyright (C) 2001-2002, Erik Andersen
+ * Copyright (C) 2001-2003, Erik Andersen
  *
  * All rights reserved.
  *
@@ -121,7 +121,7 @@ int _dl_unmap_cache(void)
 /* This function's behavior must exactly match that 
  * in uClibc/ldso/util/ldd.c */
 static struct elf_resolve * 
-search_for_named_library(char *name, int secure, const char *path_list,
+search_for_named_library(const char *name, int secure, const char *path_list,
 	struct dyn_elf **rpnt)
 {
 	int i, count = 1;
@@ -156,8 +156,7 @@ search_for_named_library(char *name, int secure, const char *path_list,
 		_dl_strcpy(mylibname, path_n); 
 		_dl_strcat(mylibname, "/"); 
 		_dl_strcat(mylibname, name);
-		if ((tpnt1 = _dl_load_elf_shared_library(secure, rpnt,
-			mylibname)) != NULL)
+		if ((tpnt1 = _dl_load_elf_shared_library(secure, rpnt, mylibname)) != NULL)
 		{
 			return tpnt1;
 		}
@@ -166,6 +165,73 @@ search_for_named_library(char *name, int secure, const char *path_list,
 	return NULL;
 }
 
+/* Check if the named library is already loaded... */
+struct elf_resolve *_dl_check_if_named_library_is_loaded(const char *full_libname)
+{
+	const char *pnt, *pnt1;
+	struct elf_resolve *tpnt1;
+	const char *libname, *libname2;
+	static const char *libc = "libc.so.";
+	static const char *aborted_wrong_lib = "%s: aborted attempt to load %s!\n";
+
+	pnt = libname = full_libname;
+
+#if defined (__SUPPORT_LD_DEBUG__)
+	if(_dl_debug) 
+		_dl_dprintf(_dl_debug_file, "Checking if '%s' is already loaded\n", full_libname);
+#endif
+	/* quick hack to ensure mylibname buffer doesn't overflow.  don't 
+	   allow full_libname or any directory to be longer than 1024. */
+	if (_dl_strlen(full_libname) > 1024)
+		return NULL;
+
+	/* Skip over any initial initial './' and '/' stuff to 
+	 * get the short form libname with no path garbage */ 
+	pnt1 = _dl_strrchr(pnt, '/');
+	if (pnt1) {
+		libname = pnt1 + 1;
+	}
+
+	/* Make sure they are not trying to load the wrong C library!
+	 * This sometimes happens esp with shared libraries when the
+	 * library path is somehow wrong! */
+#define isdigit(c)  (c >= '0' && c <= '9')
+	if ((_dl_strncmp(libname, libc, 8) == 0) &&  _dl_strlen(libname) >=8 &&
+			isdigit(libname[8]))
+	{
+		/* Abort attempts to load glibc, libc5, etc */
+		if ( libname[8]!='0') {
+			if (!_dl_trace_loaded_objects) {
+				_dl_dprintf(2, aborted_wrong_lib, libname, _dl_progname);
+				_dl_exit(1);
+			}
+			return NULL;
+		}
+	}
+
+	/* Critical step!  Weed out duplicates early to avoid
+	 * function aliasing, which wastes memory, and causes
+	 * really bad things to happen with weaks and globals. */
+	for (tpnt1 = _dl_loaded_modules; tpnt1; tpnt1 = tpnt1->next) {
+
+		/* Skip over any initial initial './' and '/' stuff to 
+		 * get the short form libname with no path garbage */ 
+		libname2 = tpnt1->libname;
+		pnt1 = _dl_strrchr(libname2, '/');
+		if (pnt1) {
+			libname2 = pnt1 + 1;
+		}
+
+		if (_dl_strcmp(libname2, libname) == 0) {
+			/* Well, that was certainly easy */
+			return tpnt1;
+		}
+	}
+
+	return NULL;
+}
+	
+
 
 /*
  * Used to return error codes back to dlopen et. al.
@@ -178,36 +244,47 @@ extern char *_dl_ldsopath;
 struct elf_resolve *_dl_load_shared_library(int secure, struct dyn_elf **rpnt,
 	struct elf_resolve *tpnt, char *full_libname)
 {
-	char *pnt;
+	char *pnt, *pnt1;
 	struct elf_resolve *tpnt1;
 	char *libname;
 
 	_dl_internal_error_number = 0;
-	pnt = libname = full_libname;
+	libname = full_libname;
 
 	/* quick hack to ensure mylibname buffer doesn't overflow.  don't 
 	   allow full_libname or any directory to be longer than 1024. */
 	if (_dl_strlen(full_libname) > 1024)
 		goto goof;
 
-	while (*pnt) {
-		if (*pnt == '/')
-			libname = pnt + 1;
-		pnt++;
+	/* Skip over any initial initial './' and '/' stuff to 
+	 * get the short form libname with no path garbage */ 
+	pnt1 = _dl_strrchr(libname, '/');
+	if (pnt1) {
+		libname = pnt1 + 1;
 	}
 
+	/* Critical step!  Weed out duplicates early to avoid
+	 * function aliasing, which wastes memory, and causes
+	 * really bad things to happen with weaks and globals. */
+	if ((tpnt1=_dl_check_if_named_library_is_loaded(libname))!=NULL)
+		return tpnt1;
+
 #if defined (__SUPPORT_LD_DEBUG__)
-	if(_dl_debug) _dl_dprintf(_dl_debug_file, "searching for library: '%s'\n", libname);
+	if(_dl_debug) _dl_dprintf(_dl_debug_file, "\tfind library='%s'; searching\n", libname);
 #endif
 	/* If the filename has any '/', try it straight and leave it at that.
 	   For IBCS2 compatibility under linux, we substitute the string 
 	   /usr/i486-sysv4/lib for /usr/lib in library names. */
 
 	if (libname != full_libname) {
+#if defined (__SUPPORT_LD_DEBUG__)
+		if(_dl_debug) _dl_dprintf(_dl_debug_file, "\ttrying file='%s'\n", full_libname);
+#endif
 		tpnt1 = _dl_load_elf_shared_library(secure, rpnt, full_libname);
-		if (tpnt1)
+		if (tpnt1) {
 			return tpnt1;
-		goto goof;
+		}
+		//goto goof;
 	}
 
 	/*
@@ -218,10 +295,9 @@ struct elf_resolve *_dl_load_shared_library(int secure, struct dyn_elf **rpnt,
 		if (tpnt->libtype == elf_executable) {
 			pnt = (char *) tpnt->dynamic_info[DT_RPATH];
 			if (pnt) {
-				pnt += (unsigned long) tpnt->loadaddr +
-					tpnt->dynamic_info[DT_STRTAB];
+				pnt += (unsigned long) tpnt->loadaddr + tpnt->dynamic_info[DT_STRTAB];
 #if defined (__SUPPORT_LD_DEBUG__)
-				if(_dl_debug) _dl_dprintf(_dl_debug_file, "searching RPATH: '%s'\n", pnt);
+				if(_dl_debug) _dl_dprintf(_dl_debug_file, "\tsearching RPATH='%s'\n", pnt);
 #endif
 				if ((tpnt1 = search_for_named_library(libname, secure, pnt, rpnt)) != NULL) 
 				{
@@ -234,7 +310,7 @@ struct elf_resolve *_dl_load_shared_library(int secure, struct dyn_elf **rpnt,
 	/* Check in LD_{ELF_}LIBRARY_PATH, if specified and allowed */
 	if (_dl_library_path) {
 #if defined (__SUPPORT_LD_DEBUG__)
-		if(_dl_debug) _dl_dprintf(_dl_debug_file, "searching _dl_library_path: '%s'\n", _dl_library_path);
+		if(_dl_debug) _dl_dprintf(_dl_debug_file, "\tsearching LD_LIBRARY_PATH='%s'\n", _dl_library_path);
 #endif
 	    if ((tpnt1 = search_for_named_library(libname, secure, _dl_library_path, rpnt)) != NULL) 
 	    {
@@ -254,6 +330,9 @@ struct elf_resolve *_dl_load_shared_library(int secure, struct dyn_elf **rpnt,
 		libentry_t *libent = (libentry_t *) & header[1];
 		char *strs = (char *) &libent[header->nlibs];
 
+#if defined (__SUPPORT_LD_DEBUG__)
+		if(_dl_debug) _dl_dprintf(_dl_debug_file, "\tsearching cache='%s'\n", LDSO_CACHE);
+#endif
 		for (i = 0; i < header->nlibs; i++) {
 			if ((libent[i].flags == LIB_ELF ||
 				 libent[i].flags == LIB_ELF_LIBC5) &&
@@ -268,7 +347,7 @@ struct elf_resolve *_dl_load_shared_library(int secure, struct dyn_elf **rpnt,
 	/* Look for libraries wherever the shared library loader
 	 * was installed */
 #if defined (__SUPPORT_LD_DEBUG__)
-	if(_dl_debug) _dl_dprintf(_dl_debug_file, "searching in ldso dir: %s\n", _dl_ldsopath);
+		if(_dl_debug) _dl_dprintf(_dl_debug_file, "\tsearching ldso dir='%s'\n", _dl_ldsopath);
 #endif
 	if ((tpnt1 = search_for_named_library(libname, secure, _dl_ldsopath, rpnt)) != NULL) 
 	{
@@ -279,14 +358,12 @@ struct elf_resolve *_dl_load_shared_library(int secure, struct dyn_elf **rpnt,
 	/* Lastly, search the standard list of paths for the library.
 	   This list must exactly match the list in uClibc/ldso/util/ldd.c */
 #if defined (__SUPPORT_LD_DEBUG__)
-	if(_dl_debug) _dl_dprintf(_dl_debug_file, "searching full lib path list\n");
+	if(_dl_debug) _dl_dprintf(_dl_debug_file, "\tsearching full lib path list\n");
 #endif
 	if ((tpnt1 = search_for_named_library(libname, secure, 
-			UCLIBC_TARGET_PREFIX "/usr/X11R6/lib:"
-			UCLIBC_TARGET_PREFIX "/usr/lib:"
-			UCLIBC_TARGET_PREFIX "/lib:"
-			UCLIBC_DEVEL_PREFIX "/lib:"
-			UCLIBC_BUILD_DIR "/lib:"
+			UCLIBC_RUNTIME_PREFIX "usr/X11R6/lib:"
+			UCLIBC_RUNTIME_PREFIX "usr/lib:"
+			UCLIBC_RUNTIME_PREFIX "lib:"
 			"/usr/lib:"
 			"/lib", rpnt)
 		    ) != NULL) 
@@ -316,12 +393,12 @@ goof:
 struct elf_resolve *_dl_load_elf_shared_library(int secure,
 	struct dyn_elf **rpnt, char *libname)
 {
-	elfhdr *epnt;
+	ElfW(Ehdr) *epnt;
 	unsigned long dynamic_addr = 0;
 	unsigned long dynamic_size = 0;
 	Elf32_Dyn *dpnt;
 	struct elf_resolve *tpnt;
-	elf_phdr *ppnt;
+	ElfW(Phdr) *ppnt;
 	char *status;
 	char header[4096];
 	unsigned long dynamic_info[24];
@@ -334,9 +411,8 @@ struct elf_resolve *_dl_load_elf_shared_library(int secure,
 	tpnt = _dl_check_hashed_files(libname);
 	if (tpnt) {
 		if (*rpnt) {
-			(*rpnt)->next = (struct dyn_elf *)
-				_dl_malloc(sizeof(struct dyn_elf));
-			_dl_memset((*rpnt)->next, 0, sizeof(*((*rpnt)->next)));
+			(*rpnt)->next = (struct dyn_elf *) _dl_malloc(sizeof(struct dyn_elf));
+			_dl_memset((*rpnt)->next, 0, sizeof(struct dyn_elf));
 			(*rpnt)->next->prev = (*rpnt);
 			*rpnt = (*rpnt)->next;
 			(*rpnt)->dyn = tpnt;
@@ -344,6 +420,9 @@ struct elf_resolve *_dl_load_elf_shared_library(int secure,
 		}
 		tpnt->usage_count++;
 		tpnt->libtype = elf_lib;
+#if defined (__SUPPORT_LD_DEBUG__)
+		if(_dl_debug) _dl_dprintf(2, "file='%s';  already loaded\n", libname);
+#endif
 		return tpnt;
 	}
 
@@ -372,7 +451,7 @@ struct elf_resolve *_dl_load_elf_shared_library(int secure,
 	}
 
 	_dl_read(infile, header, sizeof(header));
-	epnt = (elfhdr *) (intptr_t) header;
+	epnt = (ElfW(Ehdr) *) (intptr_t) header;
 	if (epnt->e_ident[0] != 0x7f ||
 		epnt->e_ident[1] != 'E' || 
 		epnt->e_ident[2] != 'L' || 
@@ -399,7 +478,7 @@ struct elf_resolve *_dl_load_elf_shared_library(int secure,
 		return NULL;
 	};
 
-	ppnt = (elf_phdr *)(intptr_t) & header[epnt->e_phoff];
+	ppnt = (ElfW(Phdr) *)(intptr_t) & header[epnt->e_phoff];
 
 	piclib = 1;
 	for (i = 0; i < epnt->e_phnum; i++) {
@@ -447,7 +526,7 @@ struct elf_resolve *_dl_load_elf_shared_library(int secure,
 	flags |= MAP_FIXED;
 
 	/* Get the memory to store the library */
-	ppnt = (elf_phdr *)(intptr_t) & header[epnt->e_phoff];
+	ppnt = (ElfW(Phdr) *)(intptr_t) & header[epnt->e_phoff];
 
 	for (i = 0; i < epnt->e_phnum; i++) {
 		if (ppnt->p_type == PT_LOAD) {
@@ -578,7 +657,7 @@ struct elf_resolve *_dl_load_elf_shared_library(int secure,
 
 	if (dynamic_info[DT_TEXTREL]) {
 #ifndef FORCE_SHAREABLE_TEXT_SEGMENTS
-		ppnt = (elf_phdr *)(intptr_t) & header[epnt->e_phoff];
+		ppnt = (ElfW(Phdr) *)(intptr_t) & header[epnt->e_phoff];
 		for (i = 0; i < epnt->e_phnum; i++, ppnt++) {
 			if (ppnt->p_type == PT_LOAD && !(ppnt->p_flags & PF_W))
 				_dl_mprotect((void *) ((piclib ? libaddr : 0) + 
@@ -595,16 +674,15 @@ struct elf_resolve *_dl_load_elf_shared_library(int secure,
 	tpnt = _dl_add_elf_hash_table(libname, (char *) libaddr, dynamic_info, 
 		dynamic_addr, dynamic_size);
 
-	tpnt->ppnt = (elf_phdr *)(intptr_t) (tpnt->loadaddr + epnt->e_phoff);
+	tpnt->ppnt = (ElfW(Phdr) *)(intptr_t) (tpnt->loadaddr + epnt->e_phoff);
 	tpnt->n_phent = epnt->e_phnum;
 
 	/*
 	 * Add this object into the symbol chain
 	 */
 	if (*rpnt) {
-		(*rpnt)->next = (struct dyn_elf *)
-			_dl_malloc(sizeof(struct dyn_elf));
-		_dl_memset((*rpnt)->next, 0, sizeof(*((*rpnt)->next)));
+		(*rpnt)->next = (struct dyn_elf *) _dl_malloc(sizeof(struct dyn_elf));
+		_dl_memset((*rpnt)->next, 0, sizeof(struct dyn_elf));
 		(*rpnt)->next->prev = (*rpnt);
 		*rpnt = (*rpnt)->next;
 		(*rpnt)->dyn = tpnt;
@@ -627,6 +705,16 @@ struct elf_resolve *_dl_load_elf_shared_library(int secure,
 		INIT_GOT(lpnt, tpnt);
 	};
 
+#if defined (__SUPPORT_LD_DEBUG__)
+	if(_dl_debug) {
+		_dl_dprintf(2, "\n\tfile='%s';  generating link map\n", libname);
+		_dl_dprintf(2, "\t\tdynamic: %x  base: %x   size: %x\n", 
+				dynamic_addr, libaddr, dynamic_size);
+		_dl_dprintf(2, "\t\t  entry: %x  phdr: %x  phnum: %d\n\n", 
+				epnt->e_entry + libaddr, tpnt->ppnt, tpnt->n_phent);
+
+	}
+#endif
 	return tpnt;
 }
 
@@ -761,7 +849,7 @@ void *_dl_malloc(int size)
 
 #if 0
 #ifdef __SUPPORT_LD_DEBUG_EARLY__
-	_dl_dprintf(_dl_debug_file, "malloc: request for %d bytes\n", size);
+	_dl_dprintf(2, "malloc: request for %d bytes\n", size);
 #endif
 #endif
 
@@ -770,7 +858,7 @@ void *_dl_malloc(int size)
 
 	if (_dl_malloc_addr - _dl_mmap_zero + size > 4096) {
 #ifdef __SUPPORT_LD_DEBUG_EARLY__
-		_dl_dprintf(_dl_debug_file, "malloc: mmapping more memory\n");
+		_dl_dprintf(2, "malloc: mmapping more memory\n");
 #endif
 		_dl_mmap_zero = _dl_malloc_addr = _dl_mmap((void *) 0, size, 
 				PROT_READ | PROT_WRITE, MAP_PRIVATE | MAP_ANONYMOUS, 0, 0);
