@@ -19,6 +19,9 @@
 
 #include <stdlib.h>
 
+extern int __drand48_iterate(unsigned short xsubi[3], 
+	struct drand48_data *buffer);
+
 int nrand48_r (xsubi, buffer, result)
      unsigned short int xsubi[3];
      struct drand48_data *buffer;
